//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "microdims.h"

ConstantRecord _ConstantRecord_instance;

typedef enum { STR_FIELD, INT_FIELD, R32_FIELD } FieldType;

typedef struct _Field
{
const char *name;
FieldType type;
int offset;
bool deduced;
} Field;

#define INSERTED(record_type,field_type,field_name) \
        { #field_name, field_type, \
          OFFSET_OF_MEMBER(record_type,field_name), false }

#define DEDUCED(record_type,field_type,field_name) \
        { #field_name, field_type, \
          OFFSET_OF_MEMBER(record_type,field_name), true }

static
const Field commonFields[]=
  {
  INSERTED(ConstantRecord,STR_FIELD,log_file_name),
  INSERTED(ConstantRecord,R32_FIELD,log_period),
  INSERTED(ConstantRecord,INT_FIELD,maximal_bacterium_count),
  INSERTED(ConstantRecord,R32_FIELD,maximal_simulation_duration),
  INSERTED(ConstantRecord,R32_FIELD,initial_glucose_concentration),
  INSERTED(ConstantRecord,STR_FIELD,acid_file_name),
  INSERTED(ConstantRecord,INT_FIELD,acid_max_step),
  INSERTED(ConstantRecord,R32_FIELD,cL_max),
  INSERTED(ConstantRecord,R32_FIELD,cA_max),
  INSERTED(ConstantRecord,R32_FIELD,cF_max),
  INSERTED(ConstantRecord,R32_FIELD,dt),
  INSERTED(ConstantRecord,R32_FIELD,detection_step),
  INSERTED(ConstantRecord,R32_FIELD,activation_step),
  INSERTED(ConstantRecord,R32_FIELD,substrate_step),
  INSERTED(ConstantRecord,R32_FIELD,substrate_size),
  INSERTED(ConstantRecord,R32_FIELD,substrate_thickness),
  INSERTED(ConstantRecord,R32_FIELD,substrate_plane_progression),
  INSERTED(ConstantRecord,R32_FIELD,substrate_CFL),
  INSERTED(ConstantRecord,INT_FIELD,substrate_force_borders),
  INSERTED(ConstantRecord,INT_FIELD,substrate_crank_nicolson),
  INSERTED(ConstantRecord,INT_FIELD,substrate_iteration_count),
  INSERTED(ConstantRecord,R32_FIELD,diffusivity_plot_width),
  INSERTED(ConstantRecord,R32_FIELD,diffusivity_plot_amount),
  INSERTED(ConstantRecord,R32_FIELD,diffusivity_plot_value),
  INSERTED(ConstantRecord,R32_FIELD,display_glucose_min),
  INSERTED(ConstantRecord,R32_FIELD,display_glucose_max),
  INSERTED(ConstantRecord,R32_FIELD,display_lactate_min),
  INSERTED(ConstantRecord,R32_FIELD,display_lactate_max),
  INSERTED(ConstantRecord,R32_FIELD,display_acetate_min),
  INSERTED(ConstantRecord,R32_FIELD,display_acetate_max),
  INSERTED(ConstantRecord,R32_FIELD,display_formate_min),
  INSERTED(ConstantRecord,R32_FIELD,display_formate_max),
  INSERTED(ConstantRecord,R32_FIELD,glucose_diffusivity),
  INSERTED(ConstantRecord,R32_FIELD,lactate_diffusivity),
  INSERTED(ConstantRecord,R32_FIELD,acetate_diffusivity),
  INSERTED(ConstantRecord,R32_FIELD,formate_diffusivity),
  DEDUCED(ConstantRecord,INT_FIELD,acid_step_count),
  DEDUCED(ConstantRecord,R32_FIELD,acid_scale),
  DEDUCED(ConstantRecord,R32_FIELD,div_cL_max),
  DEDUCED(ConstantRecord,R32_FIELD,div_cA_max),
  DEDUCED(ConstantRecord,R32_FIELD,div_cF_max),
  DEDUCED(ConstantRecord,INT_FIELD,detection_sampling),
  DEDUCED(ConstantRecord,R32_FIELD,detection_div_step),
  DEDUCED(ConstantRecord,R32_FIELD,detection_surface),
  DEDUCED(ConstantRecord,INT_FIELD,activation_tiling),
  DEDUCED(ConstantRecord,INT_FIELD,activation_sampling),
  DEDUCED(ConstantRecord,INT_FIELD,substrate_sampling),
  DEDUCED(ConstantRecord,R32_FIELD,substrate_half_size),
  DEDUCED(ConstantRecord,R32_FIELD,substrate_loop),
  DEDUCED(ConstantRecord,R32_FIELD,substrate_div_step),
  DEDUCED(ConstantRecord,R32_FIELD,substrate_cell_volume),
  DEDUCED(ConstantRecord,INT_FIELD,substrate_plane_count),
  DEDUCED(ConstantRecord,INT_FIELD,substrate_dt_divisor),
  DEDUCED(ConstantRecord,R32_FIELD,radius_min),
  DEDUCED(ConstantRecord,R32_FIELD,radius_range),
  DEDUCED(ConstantRecord,INT_FIELD,detection_cell_size),
  DEDUCED(ConstantRecord,INT_FIELD,detection_cell_capacity),
  { NULL, (FieldType)-1, 0, false }
  };

static
const Field bacteriumFields[]=
  {
  INSERTED(BacteriumConstantRecord,INT_FIELD,initial_count),
  INSERTED(BacteriumConstantRecord,R32_FIELD,distribution_center_x),
  INSERTED(BacteriumConstantRecord,R32_FIELD,distribution_center_y),
  INSERTED(BacteriumConstantRecord,R32_FIELD,distribution_inner_crown),
  INSERTED(BacteriumConstantRecord,R32_FIELD,distribution_outer_crown),
  INSERTED(BacteriumConstantRecord,R32_FIELD,display_red),
  INSERTED(BacteriumConstantRecord,R32_FIELD,display_green),
  INSERTED(BacteriumConstantRecord,R32_FIELD,display_blue),
  INSERTED(BacteriumConstantRecord,R32_FIELD,shoving_factor),
  INSERTED(BacteriumConstantRecord,R32_FIELD,bacterium_density),
  INSERTED(BacteriumConstantRecord,R32_FIELD,k_m_glucose),
  INSERTED(BacteriumConstantRecord,R32_FIELD,cP_min),
  INSERTED(BacteriumConstantRecord,R32_FIELD,q_glucose_max),
  INSERTED(BacteriumConstantRecord,R32_FIELD,lactate_acid_dissociation),
  INSERTED(BacteriumConstantRecord,R32_FIELD,acetate_acid_dissociation),
  INSERTED(BacteriumConstantRecord,R32_FIELD,formate_acid_dissociation),
  INSERTED(BacteriumConstantRecord,R32_FIELD,u_min_lactate),
  INSERTED(BacteriumConstantRecord,R32_FIELD,u_min_acetate),
  INSERTED(BacteriumConstantRecord,R32_FIELD,u_min_formate),
  INSERTED(BacteriumConstantRecord,R32_FIELD,m_pfl_glucose_ref),
  INSERTED(BacteriumConstantRecord,R32_FIELD,m_ldh_glucose_ref),
  INSERTED(BacteriumConstantRecord,R32_FIELD,q_lactate_no_growth),
  INSERTED(BacteriumConstantRecord,R32_FIELD,q_acetate_no_growth),
  INSERTED(BacteriumConstantRecord,R32_FIELD,q_formate_no_growth),
  INSERTED(BacteriumConstantRecord,R32_FIELD,y_lactate),
  INSERTED(BacteriumConstantRecord,R32_FIELD,y_acetate),
  INSERTED(BacteriumConstantRecord,R32_FIELD,y_formate),
  INSERTED(BacteriumConstantRecord,R32_FIELD,y_biomass_pfl),
  INSERTED(BacteriumConstantRecord,R32_FIELD,y_biomass_ldh),
  INSERTED(BacteriumConstantRecord,R32_FIELD,initiation_ratio_at_zero_growth),
  INSERTED(BacteriumConstantRecord,R32_FIELD,initiation_ratio_slope),
  INSERTED(BacteriumConstantRecord,R32_FIELD,replication_cycle_ps1),
  INSERTED(BacteriumConstantRecord,R32_FIELD,replication_cycle_ps2),
  INSERTED(BacteriumConstantRecord,R32_FIELD,replication_cycle_ps3),
  DEDUCED(BacteriumConstantRecord,R32_FIELD,mass_to_radius),
  { NULL, (FieldType)-1, 0, false }
  };

static
int *commonLines;
int *bacteriumLines[ASSUMED_MAX_BACTERIUM_TYPE_COUNT];

static
int // match length
parseSpaces(const char *line)
{
int pos;
for(pos=0;line[pos];++pos)
  {
  if(!isspace(line[pos]))
    {
    break;
    }
  }
return pos;
}

static
int // match length
parseIdent(const char *line,
           char *out_ident)
{
int pos=0;
if((line[pos]!='_')&&!isalpha(line[pos])) { return 0; }
out_ident[pos]=line[pos];
for(++pos;(line[pos]=='_')||isalnum(line[pos]);++pos)
  {
  out_ident[pos]=line[pos];
  }
out_ident[pos]='\0';
return pos;
}

static
int // match length
parseLine(const char *line,
          char *out_type,
          char *out_name,
          char *out_value)
{
out_type[0]='\0';
out_name[0]='\0';
out_value[0]='\0';
int pos=0;
pos+=parseSpaces(line+pos);
if(line[pos]=='{')
  {
  ++pos;
  pos+=parseSpaces(line+pos);
  int length=parseIdent(line+pos,out_type);
  if(length) { pos+=length; } else { return 0; }
  pos+=parseSpaces(line+pos);
  if(line[pos]!='}') return 0;
  ++pos;
  pos+=parseSpaces(line+pos);
  }
int length=parseIdent(line+pos,out_name);
if(length) { pos+=length; } else { return 0; }
pos+=parseSpaces(line+pos);
if(line[pos]=='=') { ++pos; } else { return 0; }
pos+=parseSpaces(line+pos);
int dest=0;
if(line[pos]=='"')
  {
  ++pos;
  out_value[dest++]='"';
  while(line[pos]&&(line[pos]!='"'))
    {
    char c=line[pos++];
    if(c=='\\')
      {
      c=line[pos++];
      switch(c)
        {
        case '0': c='\0'; break;
        case 'a': c='\a'; break;
        case 'b': c='\b'; break;
        case 't': c='\t'; break;
        case 'n': c='\n'; break;
        case 'v': c='\v'; break;
        case 'f': c='\f'; break;
        case 'r': c='\r'; break;
        }
      }
    out_value[dest++]=c;
    }
  if(line[pos]=='"') { ++pos; } else { return 0; }
  }
else
  {
  while(line[pos]&&!isspace(line[pos]))
    {
    out_value[dest++]=line[pos++];
    }
  }
out_value[dest]='\0';
pos+=parseSpaces(line+pos);
return pos;
}

static
BacteriumConstantRecord *
bacteriumRecord(const char *name)
{
ConstantRecord *cst=&_ConstantRecord_instance;
int b;
for(b=0;b<cst->bacterium_type_count;++b)
  {
  BacteriumConstantRecord *bcst=cst->bacterium+b;
  if(!strcmp(bcst->name,name))
    {
    return bcst;
    }
  }
if(b==ASSUMED_MAX_BACTERIUM_TYPE_COUNT)
  {
  DROP_DEAD("ASSUMED_MAX_BACTERIUM_TYPE_COUNT exceeded");
  }
if(b==(1<<SIZE_OF_MEMBER(Bacterium,type)))
  {
  DROP_DEAD("size of Bacterium.type exceeded");
  }
int fieldCount=(int)(sizeof(bacteriumFields)/sizeof(Field))-1;
bacteriumLines[b]=Memory_initAlloc(fieldCount,int);
cst->bacterium[b].name=Memory_allocStr(name);
++cst->bacterium_type_count;
return cst->bacterium+b;
}

static
int
findField(void *record,
          const char *name)
{
const Field *fields=record==(void*)&_ConstantRecord_instance
                    ? commonFields
                    : bacteriumFields;
for(int i=0;fields[i].name;++i)
  {
  if(!strcmp(fields[i].name,name))
    {
    return i;
    }
  }
return -1;
}

static
int *
fieldLine(void *record,
          int field)
{
ConstantRecord *cst=&_ConstantRecord_instance;
if(record==(void*)cst)
  {
  return commonLines+field;
  }
else
  {
  BacteriumConstantRecord *bcst=(BacteriumConstantRecord *)record;
  int b=(int)(bcst-cst->bacterium);
  return bacteriumLines[b]+field;
  }
}

static
bool
deducedField(void *record,
             int field)
{
return record==(void*)&_ConstantRecord_instance
       ? commonFields[field].deduced
       : bacteriumFields[field].deduced;
}

static
FieldType
fieldType(void *record,
          int field)
{
return record==(void*)&_ConstantRecord_instance
       ? commonFields[field].type
       : bacteriumFields[field].type;
}

static
intptr_t
fieldAddress(void *record,
             int field)
{
return ((intptr_t)record)+(record==(void*)&_ConstantRecord_instance
                           ? commonFields[field].offset
                           : bacteriumFields[field].offset);
}

static
char **
strField(void *record,
         int field)
{
return (char **)fieldAddress(record,field);
}

static
int *
intField(void *record,
         int field)
{
return (int *)fieldAddress(record,field);
}

static
real32_t *
r32Field(void *record,
         int field)
{
return (real32_t *)fieldAddress(record,field);
}

#define CHECK_INT_FIELD(record,field_name,unit,comp,ref) \
        do \
          { \
          int field=findField(record,#field_name); \
          int value=*intField(record,field); \
          if(!(value comp ref)) \
            { \
            Platform_msg("%s:%d: '" #field_name "' " #unit \
                         ": expecting integer " #comp " " #ref "\n", \
                         fileName,*fieldLine(record,field)); \
            ok=false; \
            } \
          } while(0)

#define CHECK_INT_FIELD2(record,field_name,unit, \
                         leftRef,leftComp, \
                         rightComp,rightRef) \
        do \
          { \
          int field=findField(record,#field_name); \
          int value=*intField(record,field); \
          if(!(leftRef leftComp value)|| \
             !(value rightComp rightRef)) \
            { \
            Platform_msg("%s:%d: '" #field_name "' " #unit \
                         ": expecting " #leftRef " " #leftComp \
                         " integer "  #rightComp " " #rightRef "\n", \
                         fileName,*fieldLine(record,field)); \
            ok=false; \
            } \
          } while(0)


#define CHECK_R32_FIELD(record,field_name,unit,comp,ref) \
        do \
          { \
          int field=findField(record,#field_name); \
          real32_t value=*r32Field(record,field); \
          if(!(value comp ref##f)) \
            { \
            Platform_msg("%s:%d: '" #field_name "' " #unit \
                         ": expecting real " #comp " " #ref "\n", \
                         fileName,*fieldLine(record,field)); \
            ok=false; \
            } \
          } while(0)

#define CHECK_R32_FIELD2(record,field_name,unit, \
                         leftRef,leftComp, \
                         rightComp,rightRef) \
        do \
          { \
          int field=findField(record,#field_name); \
          real32_t value=*r32Field(record,field); \
          if(!(leftRef##f leftComp value)|| \
             !(value rightComp rightRef##f)) \
            { \
            Platform_msg("%s:%d: '" #field_name "' " #unit \
                         ": expecting " #leftRef " " #leftComp \
                         " real "  #rightComp " " #rightRef "\n", \
                         fileName,*fieldLine(record,field)); \
            ok=false; \
            } \
          } while(0)

static
char *
expandPath(char *path,
           const char *constantFileName)
{
for(;;)
  {
  char fieldStr[0x400]="";
  const char *field=(const char *)0;
  const char *remaining=(const char *)0;
  if((field=strstr(path,"{date}")))
    {
    remaining=field+6;
    time_t nowSec=time(NULL);
    struct tm now=*localtime(&nowSec);
    sprintf(fieldStr,"%.4dm%.2dd%.2d_%.2d:%.2d:%.2d",
            1900+now.tm_year,1+now.tm_mon,now.tm_mday,
            now.tm_hour,now.tm_min,now.tm_sec);
    }
  else if((field=strstr(path,"{dir}")))
    {
    remaining=field+5;
    const char *baseName=Path_findBasename(constantFileName);
    int dirLen=(int)(baseName-constantFileName);
    memcpy(fieldStr,constantFileName,dirLen);
    fieldStr[dirLen]='\0';
    }
  else if((field=strstr(path,"{base}")))
    {
    remaining=field+6;
    strcpy(fieldStr,Path_findBasename(constantFileName));
    const char *ext=Path_findExtension(fieldStr);
    if(ext) { fieldStr[ext-fieldStr]='\0'; }
    }
  else if((field=strstr(path,"{ext}")))
    {
    remaining=field+5;
    const char *ext=Path_findExtension(constantFileName);
    if(ext) { strcpy(fieldStr,ext+1); }
    else { fieldStr[0]='\0'; }
    }
  else { break; }
  int pos=(int)(field-path);
  int remainingLen=(int)strlen(remaining);
  int fieldLen=(int)strlen(fieldStr);
  char *newPath=Memory_alloc(pos+fieldLen+remainingLen+1,char);
  memcpy(newPath,path,pos);
  memcpy(newPath+pos,fieldStr,fieldLen);
  memcpy(newPath+pos+fieldLen,remaining,remainingLen+1);
  Memory_free(path);
  path=newPath;
  }
Path_normalize(path);
return path;
}

static
char *
constantSourceCode(void *record)
{
const Field *fields=record==(void*)&_ConstantRecord_instance
                    ? commonFields : bacteriumFields;
const char *prefix=fields==commonFields ? "cst" : "bcst";
Text code;
Text_init(&code);
for(int i=0;fields[i].name;++i)
  {
  const char *name=fields[i].name;
  switch(fieldType(record,i))
    {
    case STR_FIELD:
      {
      Text_append(&code," #define %s_%s (\"%s\") \n",
                  prefix,name,*strField(record,i));
      break;
      }
    case INT_FIELD:
      {
      Text_append(&code," #define %s_%s (%d) \n",
                  prefix,name,*intField(record,i));
      break;
      }
    case R32_FIELD:
      {
      Text_append(&code," #define %s_%s (%ef) \n",
                  prefix,name,*r32Field(record,i));
      break;
      }
    }
  }
char *result=Text_stealData(&code);
Text_destroy(&code);
return result;
}

void
ConstantRecord_init(const char *fileName)
{
ConstantRecord *cst=&_ConstantRecord_instance;
int fieldCount=(int)(sizeof(commonFields)/sizeof(Field))-1;
commonLines=Memory_initAlloc(fieldCount,int);
char foundPath[PATH_MAX+1];
if(Path_lookFor(foundPath,PATH_MAX+1,fileName))
  { fileName=foundPath; }
FILE *f=Path_fopen(fileName,"r");
if(!f)
  {
  char msg[0x100];
  snprintf(msg,0x100,"\"%s\" open failure",fileName);
  SYS_DROP_DEAD(msg);
  }
int lineCount=0;
for(;;)
  {
  char line[0x400];
  if(!fgets(line,sizeof(line),f)) break;
  ++lineCount;
  char type[0x100]="";
  char name[0x100]="";
  char valueStr[0x100]="";
  int length=parseLine(line,type,name,valueStr);
  if(!length)
    {
    int skip=parseSpaces(line);
    if(line[skip]&&(line[skip]!='#')) // not empty line
      {
      Platform_msg("%s:%d:\n"
                   "  'name = value' or '{type} name = value' expected\n",
                   fileName,lineCount);
      }
    continue;
    }
  if(line[length]&&(line[length]!='#')) // not empty remaining of line
    {
    Platform_msg("%s:%d:\n"
                 "  ignoring end of line\n",
                 fileName,lineCount);
    }
  void *record=type[0] ? (void *)bacteriumRecord(type) : (void *)cst;
  int field=findField(record,name);
  if((field==-1)||deducedField(record,field))
    {
    Platform_msg("%s:%d:\n"
                 "  unknown constant '%s'\n",
                 fileName,lineCount,name);
    continue;
    }
  switch(fieldType(record,field))
    {
    case STR_FIELD:
      {
      if(valueStr[0]!='"')
        {
        Platform_msg("%s:%d:\n"
                     "  string value expected for '%s'\n",
                     fileName,lineCount,name);
        continue;
        }
      char **value=strField(record,field);
      *value=Memory_allocStr(valueStr+1);
      break;
      }
    case INT_FIELD:
      {
      int *value=intField(record,field);
      length=-1;
      sscanf(valueStr,"%d%n",value,&length);
      if((length==-1)||valueStr[length])
        {
        Platform_msg("%s:%d:\n"
                     "  integer value expected for '%s'\n",
                     fileName,lineCount,name);
        continue;
        }
      break;
      }
    case R32_FIELD:
      {
      real32_t *value=r32Field(record,field);
      length=-1;
      sscanf(valueStr,"%f%n",value,&length);
      if((length==-1)||valueStr[length])
        {
        Platform_msg("%s:%d:\n"
                     "  real value expected for '%s'\n",
                     fileName,lineCount,name);
        continue;
        }
      break;
      }
    }
  int *defLine=fieldLine(record,field);
  if(*defLine)
    {
    Platform_msg("%s:%d:\n"
                 "  overriding '%s' previously set on line %d\n",
                 fileName,lineCount,name,*defLine);
    }
  *defLine=lineCount;
  }
fclose(f);
bool ok=true;
for(int i=0;commonFields[i].name;++i)
  {
  if(!commonLines[i]&&!commonFields[i].deduced)
    {
    Platform_msg("%s: missing definition of '%s' constant\n",
                 fileName,
                 commonFields[i].name);
    ok=false;
    }
  }
for(int b=0;b<cst->bacterium_type_count;++b)
  {
  for(int i=0;bacteriumFields[i].name;++i)
    {
    if(!bacteriumLines[b][i]&&!bacteriumFields[i].deduced)
      {
      Platform_msg("%s: missing definition of '{%s} %s' constant\n",
                   fileName,
                   cst->bacterium[b].name,
                   bacteriumFields[i].name);
      ok=false;
      }
    }
  }

//-- simulation specific parameters --
// log_file_name
CHECK_R32_FIELD(cst,log_period,[s],>,0.0);
CHECK_INT_FIELD(cst,maximal_bacterium_count,[1],>=,0);
CHECK_R32_FIELD(cst,maximal_simulation_duration,[s],>=,0.0);
CHECK_R32_FIELD(cst,initial_glucose_concentration,[MG/l],>=,0.0);
// acid_file_name
CHECK_INT_FIELD(cst,acid_max_step,[1],>,0);
CHECK_R32_FIELD(cst,cL_max,[MA/l],>=,0.0);
CHECK_R32_FIELD(cst,cA_max,[MA/l],>=,0.0);
CHECK_R32_FIELD(cst,cF_max,[MA/l],>=,0.0);
CHECK_R32_FIELD(cst,dt,[s],>,0.0);
CHECK_R32_FIELD(cst,detection_step,[m],>,0.0);
CHECK_R32_FIELD(cst,activation_step,[m],>,0.0);
CHECK_R32_FIELD(cst,substrate_step,[m],>,0.0);
CHECK_R32_FIELD(cst,substrate_size,[m],>,0.0);
CHECK_R32_FIELD(cst,substrate_thickness,[m],>,0.0);
CHECK_R32_FIELD(cst,substrate_plane_progression,[1],>=,1.0);
CHECK_R32_FIELD(cst,substrate_CFL,[1],>=,0.0);
CHECK_INT_FIELD2(cst,substrate_force_borders,[1],0,<=,<=,1);
CHECK_INT_FIELD2(cst,substrate_crank_nicolson,[1],0,<=,<=,1);
CHECK_INT_FIELD(cst,substrate_iteration_count,[1],>,1);
CHECK_R32_FIELD(cst,diffusivity_plot_width,[m],>=,0.0);
CHECK_R32_FIELD2(cst,diffusivity_plot_amount,[m],0.0,<=,<=,1.0);
CHECK_R32_FIELD(cst,diffusivity_plot_value,[m],>=,0.0);

//-- display specific parameters --
// display_glucose_min
// display_glucose_max
// display_lactate_min
// display_lactate_max
// display_acetate_min
// display_acetate_max
// display_formate_min
// display_formate_max

//-- model specific parameters for environment --
CHECK_R32_FIELD(cst,glucose_diffusivity,[m^2/s],>,0.0);
CHECK_R32_FIELD(cst,lactate_diffusivity,[m^2/s],>,0.0);
CHECK_R32_FIELD(cst,acetate_diffusivity,[m^2/s],>,0.0);
CHECK_R32_FIELD(cst,formate_diffusivity,[m^2/s],>,0.0);

//-- deduced from above --
cst->log_file_name=expandPath(cst->log_file_name,fileName);
cst->acid_step_count=cst->acid_max_step+1;
cst->acid_scale=(real32_t)cst->acid_max_step;
cst->div_cL_max=1.0f/cst->cL_max;
cst->div_cA_max=1.0f/cst->cA_max;
cst->div_cF_max=1.0f/cst->cF_max;
cst->detection_sampling=(int)fmaxf(1.0f,roundf(
                             cst->substrate_size/cst->detection_step));
cst->detection_step=cst->substrate_size/cst->detection_sampling;
cst->detection_div_step=0.999999f/cst->detection_step;
cst->detection_surface=cst->detection_step*cst->detection_step;
cst->activation_tiling=(int)fmaxf(1.0f,roundf(
                            cst->activation_step/cst->detection_step));
cst->activation_step=cst->detection_step*cst->activation_tiling;
cst->activation_sampling=(int)(Platform_alignSize(cst->detection_sampling,
                                                  cst->activation_tiling)/
                               cst->activation_tiling);
cst->substrate_sampling=(int)fmaxf(1.0f,roundf(
                             cst->substrate_size/cst->substrate_step));
cst->substrate_step=cst->substrate_size/cst->substrate_sampling;
cst->substrate_half_size=0.5f*cst->substrate_size;
cst->substrate_loop=cst->substrate_size-0.001f*cst->detection_step;
cst->substrate_div_step=0.999999f/cst->substrate_step;
cst->substrate_cell_volume=1e3f*cst->substrate_step*
                                cst->substrate_step*
                                cst->substrate_step;
cst->substrate_plane_count=1;
real32_t step=cst->substrate_step;
real32_t thickness=step;
while(thickness<cst->substrate_thickness)
  {
  ++cst->substrate_plane_count;
  thickness+=(step*=cst->substrate_plane_progression);
  }
real32_t pMin=1.0f;
real32_t pMax=cst->substrate_plane_progression;
while(pMax-pMin>1e-6f)
  {
  const real32_t testP=0.5f*(pMin+pMax);
  real32_t testS=cst->substrate_step;
  real32_t testT=testS;
  for(int i=1,i_end=cst->substrate_plane_count;i<i_end;++i)
    {
    testT+=(testS*=testP);
    }
  if(testT>cst->substrate_thickness)
    {
    pMax=testP;
    }
  else
    {
    pMin=testP;
    }
  }
cst->substrate_plane_progression=0.5f*(pMin+pMax);
step=cst->substrate_step;
cst->substrate_thickness=step;
for(int i=1,i_end=cst->substrate_plane_count;i<i_end;++i)
  {
  cst->substrate_thickness+=(step*=cst->substrate_plane_progression);
  }
if(cst->substrate_CFL>0.0f)
  {
  const real32_t maxDiff=fmaxf(fmaxf(cst->glucose_diffusivity,
                                     cst->lactate_diffusivity),
                               fmaxf(cst->acetate_diffusivity,
                                     cst->formate_diffusivity));
  const real32_t maxDt=cst->substrate_CFL*
                       cst->substrate_step*cst->substrate_step/maxDiff;
  cst->substrate_dt_divisor=(int)ceilf(cst->dt/maxDt);
  }
else
  {
  cst->substrate_dt_divisor=1;
  }

//-- bacterium specific parameters --
real32_t radius_min=FLT_MAX;
real32_t radius_max=0.0f;
real32_t radius_occ=FLT_MAX;
for(int b=0;b<cst->bacterium_type_count;++b)
  {
  BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;

  //-- simulation specific parameters --
  CHECK_INT_FIELD(bcst,initial_count,[1],>=,0);
  CHECK_R32_FIELD2(bcst,distribution_center_x,[1],-1.0,<=,<=,1.0);
  CHECK_R32_FIELD2(bcst,distribution_center_y,[1],-1.0,<=,<=,1.0);
  CHECK_R32_FIELD(bcst,distribution_inner_crown,[1],>=,0.0);
  CHECK_R32_FIELD(bcst,distribution_outer_crown,[1],>=,0.0);

  //-- display specific parameters --
  CHECK_R32_FIELD2(bcst,display_red,[1],0.0,<=,<=,1.0);
  CHECK_R32_FIELD2(bcst,display_green,[1],0.0,<=,<=,1.0);
  CHECK_R32_FIELD2(bcst,display_blue,[1],0.0,<=,<=,1.0);

  //-- model specific parameters --
  CHECK_R32_FIELD(bcst,shoving_factor,[1],>,0.0);
  CHECK_R32_FIELD(bcst,bacterium_density,[gDW/m^3],>,0.0);
  CHECK_R32_FIELD(bcst,cP_min,[MP/l],>,0.0);
  CHECK_R32_FIELD(bcst,initiation_ratio_at_zero_growth,[gDW],>=,0.0);
  CHECK_R32_FIELD(bcst,initiation_ratio_slope,[gDW.s],<=,0.0);
  CHECK_R32_FIELD(bcst,q_glucose_max,[MG/(gDW.s)],>=,0.0);
  CHECK_R32_FIELD(bcst,k_m_glucose,[MG/l],>=,0.0);
  CHECK_R32_FIELD(bcst,lactate_acid_dissociation,[MP/l],>,0.0);
  CHECK_R32_FIELD(bcst,acetate_acid_dissociation,[MP/l],>,0.0);
  CHECK_R32_FIELD(bcst,formate_acid_dissociation,[MP/l],>,0.0);
  CHECK_R32_FIELD(bcst,u_min_lactate,[MUA/l],>,0.0);
  CHECK_R32_FIELD(bcst,u_min_acetate,[MUA/l],>,0.0);
  CHECK_R32_FIELD(bcst,u_min_formate,[MUA/l],>,0.0);
  CHECK_R32_FIELD(bcst,m_pfl_glucose_ref,[MG/(gDW.s)],>,0.0);
  CHECK_R32_FIELD(bcst,m_ldh_glucose_ref,[MG/(gDW.s)],>,0.0);
  CHECK_R32_FIELD(bcst,q_lactate_no_growth,[MA/(gDW.s)],>=,0.0);
  CHECK_R32_FIELD(bcst,q_acetate_no_growth,[MA/(gDW.s)],>=,0.0);
  CHECK_R32_FIELD(bcst,q_formate_no_growth,[MA/(gDW.s)],>=,0.0);
  CHECK_R32_FIELD(bcst,y_lactate,[MA/MG],>=,0.0);
  CHECK_R32_FIELD(bcst,y_acetate,[MA/MG],>=,0.0);
  CHECK_R32_FIELD(bcst,y_formate,[MA/MG],>=,0.0);
  CHECK_R32_FIELD(bcst,y_biomass_pfl,[gDW/MG],>=,0.0);
  CHECK_R32_FIELD(bcst,y_biomass_ldh,[gDW/MG],>=,0.0);
  CHECK_R32_FIELD(bcst,replication_cycle_ps1,[1/s],>,0.0);
  CHECK_R32_FIELD(bcst,replication_cycle_ps2,[1],<,0.0);
  CHECK_R32_FIELD(bcst,replication_cycle_ps3,[s],>,0.0);

  //-- deduced from above --
  bcst->mass_to_radius=3.0f/(bcst->bacterium_density*4.0f*M_PIf);

  // FIXME: what are the correct estimates for min, max and average radius?
  radius_min=fminf(radius_min,Bacterium_radiusFromMass(b,
                              0.5f*bcst->initiation_ratio_at_zero_growth));
  radius_max=fmaxf(radius_max,Bacterium_radiusFromMass(b,
                              2.0f*bcst->initiation_ratio_at_zero_growth));
  radius_occ=fminf(radius_occ,0.5f*(1.0f+bcst->shoving_factor)*
                              Bacterium_radiusFromMass(b,
                              1.0f*bcst->initiation_ratio_at_zero_growth));

  Memory_free(bacteriumLines[b]);
  }
Memory_free(commonLines);
if(!ok)
  {
  DROP_DEAD("incorrect constant file");
  }
cst->radius_min=radius_min;
cst->radius_range=radius_max-radius_min;
const int minCellCapacity=
  (int)ceilf(cst->detection_step*cst->detection_step/
             Bacterium_cellOccupancyFromRadius(radius_occ));
cst->detection_cell_size=(int)Platform_alignSize(
                              OFFSET_OF_MEMBER(DetectionCell,content)+
                              minCellCapacity*sizeof(Bacterium),
                              ASSUMED_CACHE_ALIGNMENT);
cst->detection_cell_capacity=(int)((cst->detection_cell_size-
                                    OFFSET_OF_MEMBER(DetectionCell,content))/
                                   sizeof(Bacterium));
#if 0
  Platform_msg("sizeof(Bacterium)=%d\n",(int)sizeof(Bacterium));
  Platform_msg("detection_cell_size=%d\n",cst->detection_cell_size);
  Platform_msg("minCellCapacity=%d\n",minCellCapacity);
  Platform_msg("detection_cell_capacity=%d\n",cst->detection_cell_capacity);
#endif

cst->acid_file_name=expandPath(cst->acid_file_name,fileName);
Platform_msg("loading proton concentrations from \"%s\"\n",
             cst->acid_file_name);
f=Path_fopen(cst->acid_file_name,"r");
if(!f)
  {
  char msg[0x100];
  snprintf(msg,0x100,"\"%s\" open failure",cst->acid_file_name);
  SYS_DROP_DEAD(msg);
  }
const intptr_t acidRecordCount=cst->acid_step_count*
                               cst->acid_step_count*
                               cst->acid_step_count;
real32_t * RESTRICT cP_array=Memory_alloc(acidRecordCount,real32_t);
for(intptr_t i=0;i<acidRecordCount;++i)
  {
  if(fscanf(f," %g ",cP_array+i)!=1)
    { DROP_DEAD("acid file read failure"); }
  }
fclose(f);
for(int b=0;b<cst->bacterium_type_count;++b)
  {
  BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
  bcst->acidRecords=Memory_alloc(acidRecordCount,AcidRecord);
  for(intptr_t i=0;i<acidRecordCount;++i)
    {
    // for(cL ...) for(cA ...) for(cF ...) ...
    const real32_t cP=cP_array[i];

    const real32_t cP_cPmin=(cP-1e-7f)/(bcst->cP_min-1e-7f); // [1]
    const real32_t pH=-log10f(cP); // [1]
    const real32_t LDH_fraction=1.0f/(1.0f+expf(1.9547f*(pH-5.7809f))); // [1]
    const real32_t PFL_fraction=1.0f-LDH_fraction; // [1]
    const real32_t F_fraction=1.0f-
                              1.0f/(1.0f+expf(11.3021f*(pH-6.3133f))); // [1]

    const intptr_t iL=i/(cst->acid_step_count*cst->acid_step_count);
    const intptr_t iA=(i/cst->acid_step_count)%cst->acid_step_count;
    const intptr_t iF=i%cst->acid_step_count;
    const real32_t cL=cst->cL_max*powf(iL/cst->acid_scale,6.0f); // [MA/l]
    const real32_t cA=cst->cA_max*powf(iA/cst->acid_scale,6.0f); // [MA/l]
    const real32_t cF=cst->cF_max*powf(iF/cst->acid_scale,6.0f); // [MA/l]

    // concentrations of undissociated weak acids [MUA/l]
    const real32_t uL=(cL*cP)/(bcst->lactate_acid_dissociation+cP);
    const real32_t uA=(cA*cP)/(bcst->acetate_acid_dissociation+cP);
    const real32_t uF=(cF*cP)/(bcst->formate_acid_dissociation+cP);

    // coefficient of [H+] maintenance term for PFL and LDH metabolism,
    // [MG/(gDW.s)]
    const real32_t a_pfl=bcst->q_glucose_max-bcst->m_pfl_glucose_ref;
    const real32_t a_ldh=bcst->q_glucose_max-bcst->m_ldh_glucose_ref;

    // coefficient of weak acid maintenance terms for PFL and LDH metabolisms,
    // [MG/(gDW.s)] , 10**-4.2 = 6.3096e-5 , 10**-5.4 = 3.9811e-6
    const real32_t b_a_factor=1.0f-(6.3096e-5f+3.9811e-6f-2.0f*1e-7f)/
                                   (2.0f*(bcst->cP_min-1e-7f));
    const real32_t b_pfl=a_pfl*b_a_factor;
    const real32_t b_ldh=a_ldh*b_a_factor;

    // PFL and LDH maintenance coefficients // [MG/(gDW.s)]
    const real32_t sigma_u_uMin=uA/bcst->u_min_acetate+
                                uF/bcst->u_min_formate+
                                uL/bcst->u_min_lactate; // [1]
    const real32_t maintCoef_PFL=bcst->m_pfl_glucose_ref+
                                 a_pfl*cP_cPmin+
                                 b_pfl*sigma_u_uMin; // [MG/(gDW.s)]
    const real32_t maintCoef_LDH=bcst->m_ldh_glucose_ref+
                                 a_ldh*cP_cPmin+
                                 b_ldh*sigma_u_uMin; // [MG/(gDW.s)]

    // maintenance and biomass yield coefficient as a function of pH
    const real32_t maintCoef=PFL_fraction*maintCoef_PFL+
                             LDH_fraction*maintCoef_LDH; // [MG/(gDW.s)]
    const real32_t massYieldCoef=PFL_fraction*bcst->y_biomass_pfl+
                                 LDH_fraction*bcst->y_biomass_ldh; // [gDW/MG]

    AcidRecord *ar=bcst->acidRecords+i;
    ar->maintCoef=maintCoef;
    ar->massYieldCoef=massYieldCoef;
    ar->qL_glucose=LDH_fraction*bcst->y_lactate;
    ar->qL_maintenance=LDH_fraction*maintCoef_LDH*
                       (bcst->q_lactate_no_growth/bcst->m_ldh_glucose_ref);
    ar->qA_glucose=PFL_fraction*bcst->y_acetate;
    ar->qA_maintenance=PFL_fraction*maintCoef_PFL*
                       (bcst->q_acetate_no_growth/bcst->m_pfl_glucose_ref);
    ar->qF_glucose=PFL_fraction*F_fraction*bcst->y_formate;
    ar->qF_maintenance=PFL_fraction*F_fraction*maintCoef_PFL*
                       (bcst->q_formate_no_growth/bcst->m_pfl_glucose_ref);
    }
  }
Memory_free(cP_array);

cst->constantSourceCode=constantSourceCode(cst);
for(int b=0;b<cst->bacterium_type_count;++b)
  {
  BacteriumConstantRecord * RESTRICT bcst=cst->bacterium+b;
  bcst->constantSourceCode=constantSourceCode(bcst);
  const char *sourceFileName="bacterium.dyn_c";
  Text code;
  Text_init(&code);
  Text_appendStr(&code,
    "                          \n"
    " #include \"microdims.h\" \n"
    "                          \n");
  Text_appendStr(&code,
    cst->constantSourceCode);
  Text_appendStr(&code,
    " \n");
  Text_appendStr(&code,
    bcst->constantSourceCode);
  Text_append(&code,
    "                \n"
    " #line 1 \"%s\" \n",
    sourceFileName);
  intptr_t fileSize=(intptr_t)Path_fileSize(sourceFileName);
  if(fileSize>0)
    {
    intptr_t len=Text_length(&code);
    char *t=Text_get(&code)+len;
    Text_setLength(&code,len+fileSize);
    if(Path_readFile(sourceFileName,t,fileSize,0)!=fileSize)
      {
      fileSize=-1;
      }
    }
  if(fileSize<0)
    {
    Platform_msg("cannot load file %s\n",sourceFileName);
    DROP_DEAD("dynamic code failure\n");
    }
  DynamicCompiler dc;
  DynamicCompiler_init(&dc);
  bcst->dynamicCode=DynamicCompiler_build(&dc,Text_get(&code));
  if(DynamicCompiler_buildLog(&dc))
    {
    Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    Platform_msg("%s\n",DynamicCompiler_buildLog(&dc));
    Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    }
  if(!bcst->dynamicCode)
    {
    DROP_DEAD("code compilation failure\n");
    }
  Text_destroy(&code);
  Plugin_function(&bcst->creationFnct,
                  bcst->dynamicCode,"Bacterium_create");
  if(!bcst->creationFnct)
    {
    char msg[0x100];
    Plugin_lastErrorMsg(msg,0x100);
    Platform_msg("Plugin_function() failure: %s\n",msg);
    DROP_DEAD("dynamic code failure\n");
    }
  Plugin_function(&bcst->activationFnct,
                  bcst->dynamicCode,"Bacterium_activate");
  if(!bcst->activationFnct)
    {
    char msg[0x100];
    Plugin_lastErrorMsg(msg,0x100);
    Platform_msg("Plugin_function() failure: %s\n",msg);
    DROP_DEAD("dynamic code failure\n");
    }
  DynamicCompiler_destroy(&dc);
  }
}

//----------------------------------------------------------------------------
